/*
 * VARiD: A Variation Detection Framework for Colorspace and Letterspace platforms
 * Contact: <varid at cs dot toronto dot edu>
 * Copyright 2010 the authors (see README).
 *
 * This file is part of VARiD.
 *
 * VARiD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * VARiD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VARiD.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include <math.h>
#include <float.h>
#include "va_assert.h"
#include "va_log.h"
#include "va_time.h"
#include "varid.h"



VA_DOUBLE varid_viterbi(
    VA_MatDouble   trans_mat, 
    CData_t        cdata, 
    VA_MatDouble   e, 
    Maps_t         maps, 
    VA_VecI32      window,
    VA_MatDouble * viterbi_prob)
{
    struct timespec ts;
    struct timespec te;
    struct timespec tdiff;
    int rc, i, j;
    size_t sz = cdata.ref.size;


    rc = va_GetTime(&ts);
    VA_ASSERT(rc == 0);

    //Viterbi
    VA_ASSERT(viterbi_prob!=NULL);
    VA_MatI32 viterbi_backtrace = va_allocMatI32(sz,trans_mat.size_y);
    VA_ASSERT(viterbi_backtrace.data!=NULL);
    //VA_MatU8 *viterbi_haplomes = va_allocMatU8(sz,2);
    //VA_ASSERT(viterbi_haplomes!=NULL);

    VA_VecDouble s = va_allocVecDouble(sz);
    for (i=0; i<sz; i++) {
    	s.data[i]=-1.0/0.0;
    }

    int start = window.data[0]-1; int end = window.data[1];
    //get the begining in
    for (i=0; i<40; i++) {
	for (j=0; j<40; j++) {
		int state = i*40 + j;
		if (i<4 && j<4) {
			//f->data[state*f->size_x+start]=e.data[start*1600+state]-log(1600);
			viterbi_prob->data[start*1600 + state]=e.data[start*1600+state]-log(1600);
			s.data[start]=MAX(s.data[start],viterbi_prob->data[start*1600 + state]);
		} else {
			//f->data[state*f->size_x+start]=-FLT_MAX;
			viterbi_prob->data[start*1600 + state]=-1.0/0.0;
		}
	}
    }
    for (i=0; i<1600; i++) {
    	viterbi_prob->data[start*1600+i]-=s.data[start];
    }
    int t;
    for (t=start+1; t<end; t++) {
	#pragma omp parallel
	{
	int to_state; double largest_value=-1.0/0.0;
	#pragma omp for
	for (to_state=0; to_state<1600; to_state++) {
		//compute the f values for this position
		double max_log_value=-1.0/0.0;
		int max_previous_state=-1;
		int i;
		for (i=0; i<incoming_used[to_state]; i++) {
			int from_state=incoming[to_state][i];
			if (isinf(trans_mat.data[ from_state * 1600 + to_state ])) {
				fprintf(stderr,"A fatal error has occured in the Viterbi alogorithm optimizations\n");
				exit(1);
			}
			//summands[i]=f->data[ from_state * f->size_x + (t-1)] + trans_mat.data[ from_state * 1600 + to_state];
			double log_value=viterbi_prob->data[(t-1)*1600 + from_state] +  e.data[t*1600 + to_state] + trans_mat.data[ from_state * 1600 + to_state];
			if (log_value>max_log_value) {
				max_previous_state=from_state;
				max_log_value=log_value;
			}
		}
		viterbi_backtrace.data[(t-1)*1600 + to_state]=max_previous_state;
		viterbi_prob->data[t*1600+to_state]=max_log_value;
		//s.data[t]=MAX(s.data[t],viterbi_prob->data[t*1600+to_state]);
		largest_value=MAX(largest_value,viterbi_prob->data[t*1600+to_state]);
	}
	#pragma omp critical
	{
		if (s.data[t]<largest_value) {
			s.data[t]=largest_value;
		}
	}	
	}
	for (i=0; i<1600; i++) {
		viterbi_prob->data[t*1600+i]-=s.data[t];
	} 
	
    }
    //for (i=0; i<sz; i++) {
    //viterbi_haplomes->data[i]='N';
    //viterbi_haplomes->data[sz+i]='N';
    //}
    int state; double max_value=-1.0/0.0; int max_state=-1;
    for (state=0; state<1600; state++ ){ 
	if (viterbi_prob->data[(end-1)*1600+state]>max_value) {
		max_value=viterbi_prob->data[(end-1)*1600+state];
		max_state=state;
	}
    }
    memset(viterbi_prob->data,0,sizeof(VA_DOUBLE)*(viterbi_prob->size_x*viterbi_prob->size_y));
    if (max_state<0 || max_state>=1600){ 
	fprintf(stderr,"A fatal error has occured in backtracing of viterbi\n");
	exit(1);
    }
    viterbi_prob->data[(end-1)*1600+max_state]=1.00;
    //viterbi_haplomes->data[end-1]=state_list[max_state/40][1];
    //viterbi_haplomes->data[sz+end-1]=state_list[max_state%40][1];
    int backtrace_state=max_state;
    for (t=end-1-1; t>=0; t--) {
	backtrace_state=viterbi_backtrace.data[t*1600+backtrace_state];
	if (backtrace_state>=1600 || backtrace_state<0) {
		fprintf(stderr,"A fatal error has occured in backtrace state of viterbi\n");
		exit(1);
	}
	viterbi_prob->data[t*1600+backtrace_state]=1.00;
	//viterbi_haplomes->data[t]=state_list[backtrace_state/40][1];
	//viterbi_haplomes->data[sz+t]=state_list[backtrace_state%40][1];
    }
    /*for (i=0; i<sz; i++) {
	char hap1 = viterbi_haplomes->data[i];
	char hap2 = viterbi_haplomes->data[sz+i];
	char ref;
	switch (cdata.ref.data[i]) {
		case 0:
			ref='a';
			break;
		case 1:
			ref='c';
			break;
		case 2:
			ref='g';
			break;
		case 3:
			ref='t';
			break;
		default:
			ref='X';
			break;
	}
	char type;
	if (hap1==hap2 && hap1==ref) {
		type='R';
	} else if (hap1==hap2) {
		type='A';
	} else {
		type='H';
	}
	printf("%6d\t%c/%c\t%c\n",i,hap1,hap2,type);
    }
    VA_VecU8 temp; temp.data=viterbi_haplomes->data; temp.size=sz;
    temp=nuc2nuc(temp);
    memcpy(viterbi_haplomes->data,temp.data,sz*sizeof(VA_UINT8));
    free(temp.data);
    temp.data=viterbi_haplomes->data+sz; temp.size=sz;
    temp=nuc2nuc(temp);
    memcpy(viterbi_haplomes->data+sz,temp.data,sz*sizeof(VA_UINT8));
    free(temp.data);*/

    free(viterbi_backtrace.data);
    /* print time stats */
    rc = va_GetTime(&te);
    VA_ASSERT(rc == 0);

    tdiff = va_TimeSub(&te, &ts);
    fprintf(stderr, "\t\tviterbi algo done in: %ld.%03ld secs\n",
        tdiff.tv_sec, tdiff.tv_nsec/NS_IN_MS);
 

   //for(i=0; i<20; i++) {
	//printf("%3d %s\n",i,state_string(maps.full_letter.data,i));
   //}

    return 0;


}
