/*
 * VARiD: A Variation Detection Framework for Colorspace and Letterspace platforms
 * Contact: <varid at cs dot toronto dot edu>
 * Copyright 2010 the authors (see README).
 *
 * This file is part of VARiD.
 *
 * VARiD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * VARiD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VARiD.  If not, see <http://www.gnu.org/licenses/>.
 */

/******************************************************************************
* These macros provide methods for simple (value) confirmations as well as 
* logging and program termination should the supplied expression not be true.
*
* A knowledgeable reader will note that this header does not meet the
* idempotence requirements established for "assert.h" by the ANSI C standard.
*******************************************************************************/
#ifndef VA_ASSERT_H
#define VA_ASSERT_H

#include "va_log.h"
#include <stdlib.h>
#include <errno.h>  

/* Determine if any code is to be substituted for ASSERT */
#ifndef NDEBUG

/* Define macro to produce a message and abort the process if the
 * supplied "assertion" is not true.
 */


#  define VA_ASSERT(_a)                                                       \
   do                                                                         \
   {                                                                          \
      if (!(_a))                                                              \
      {                                                                       \
         LOG_FATAL1("assertion: (" #_a ") failed, Last errno=%d", errno);     \
         exit (EXIT_FAILURE);                                                 \
      }                                                                       \
   } while(0)
 
#else    /* NDEBUG */

#  define VA_ASSERT(_a)	 ((void)0)
 
#endif   /* NDEBUG */

/* Define macro to produce a "fatal" message and exit the process if the
 * supplied "affirmation" is not true.
 */
#define VA_AFFIRM(_a)                                                          \
   do                                                                          \
   {                                                                           \
      if (!(_a))                                                               \
      {                                                                        \
         LOG_FATAL0("affirmation: (" #_a ") failed");                          \
         exit (EXIT_FAILURE);                                                  \
      }                                                                        \
   }                                                                           \
   while (0)

#endif   /* VA_ASSERT_H */
