/*
 * VARiD: A Variation Detection Framework for Colorspace and Letterspace platforms
 * Contact: <varid at cs dot toronto dot edu>
 * Copyright 2010 the authors (see README).
 *
 * This file is part of VARiD.
 *
 * VARiD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * VARiD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VARiD.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include <math.h>
#include <time.h>
#include "config.h"
#include "va_assert.h"
#include "va_log.h"
#include "va_time.h"
#include "varid.h"

void writeVCF(
    VA_UINT32    startpos, 
    const char * ref_name, 
    VA_VecU8     hap1, 
    VA_VecU8     hap2, 
    VA_VecU8     ref, 
    VA_MatDouble maxvals_2,
    Covmaps_t    covmap,
    VA_MatU8 cmat,
    VA_MatU8 lmat)
{
    int h1t = TRUE;
    int h2t = TRUE;
    int h3t = TRUE;
    VA_VecU8 don1;
    VA_VecU8 don2;
    VA_VecU8 ref2;
    int i;
    int pos;

    /* transform the data to ls format (a,c,g,t,-,n) */
    for (i = 0; i < hap1.size; i++) {
        if (isalnum(hap1.data[i]) == 0)
            h1t = FALSE;
        if (isalnum(hap2.data[i]) == 0)
            h2t = FALSE;
        if (isalnum(ref.data[i]) == 0)
            h3t = FALSE;
    }
        
    VA_ASSERT(h1t == h2t && h1t == h3t);
    if (!h1t) {
        don1 = nuc2nuc(hap1);
        don2 = nuc2nuc(hap2);
        ref2 = nuc2nuc(ref);
    } else {
        don1 = va_allocVecU8(hap1.size);
        don2 = va_allocVecU8(hap1.size);
        ref2 = va_allocVecU8(ref.size);

        don1 = hap1;
        don2 = hap2;
        memcpy(don1.data, hap1.data, sizeof(VA_UINT8) * don1.size);
        memcpy(don2.data, hap2.data, sizeof(VA_UINT8) * don2.size);
        memcpy(ref2.data, ref.data, sizeof(VA_UINT8) * ref2.size);
    }
    
    //clean up
    VA_ASSERT(don1.size==ref.size);
    VA_ASSERT(don2.size==ref.size);
    VA_VecU8 ref2_new = va_allocVecU8(ref.size);
    VA_VecU8 don1_new = va_allocVecU8(don1.size);
    VA_VecU8 don2_new = va_allocVecU8(don2.size);
    VA_VecDouble confidence = va_allocVecDouble(ref.size);
    VA_VecU8 cmat1 = va_allocVecU8(ref.size);
    VA_VecU8 cmat2 = va_allocVecU8(ref.size);
    VA_VecU8 lmat1 = va_allocVecU8(ref.size);
    VA_VecU8 lmat2 = va_allocVecU8(ref.size);
    VA_VecU32 lookup = va_allocVecU32(ref.size);
    int index=0;
    for (i=0; i<ref2.size; i++) {
	if (don1.data[i]!='-' || don2.data[i] != '-' || ref2.data[i]!='-'){
		don1_new.data[index]=don1.data[i];
		don2_new.data[index]=don2.data[i];
		confidence.data[index]=maxvals_2.data[i];
		cmat1.data[index]=cmat.data[i];
		cmat2.data[index]=cmat.data[cmat.size_x + i];
		lmat1.data[index]=lmat.data[i];
		lmat2.data[index]=lmat.data[lmat.size_x + i];
		lookup.data[index]=i;
		ref2_new.data[index++]=ref2.data[i];
	}
    }
    free(ref2.data);
    ref2 = ref2_new;
    ref2.size=index;
    free(don1.data);
    don1 = don1_new;
    VA_ASSERT(don1.size>=index);
    don1.size=index;
    free(don2.data);
    don2 = don2_new;
    VA_ASSERT(don2.size>=index);
    don2.size=index;
    cmat1.size=index;
    cmat2.size=index;
    lmat1.size=index;
    lmat2.size=index;
    lookup.size=index;



    /*for (i=0; i<lmat1.size; i++ ){
	fprintf(stderr,"%d",lmat1.data[i]);
    }
    fprintf(stderr,"\n");
    for (i=0; i<cmat1.size; i++ ){
	fprintf(stderr,"%d",cmat1.data[i]);
    }
    fprintf(stderr,"\n");
    for (i=0; i<lmat2.size; i++ ){
	fprintf(stderr,"%d",lmat2.data[i]);
    }
    fprintf(stderr,"\n");
    for (i=0; i<cmat2.size; i++ ){
	fprintf(stderr,"%d",cmat2.data[i]);
    }
    fprintf(stderr,"\n");
    for (i=0; i<lookup.size; i++ ){
	fprintf(stderr,"%d ",lookup.data[i]);
    }
    fprintf(stderr,"\n");*/
    //print the VCF header
    if (no_header==0) {
	    fprintf(output_file,"##fileformat=VCFv4.0\n");
	    char date[256];
	    time_t rawtime;
	    time ( &rawtime );
	    struct tm * timeinfo = localtime ( &rawtime );
	    strftime(date,256,"%Y%m%d",timeinfo);
	    fprintf(output_file,"##fileDate=%s\n",date);
	    fprintf(output_file,"##source=%s\n",PACKAGE_STRING);
	    fprintf(output_file,"##reference=%s\n",ref_name);
	    fprintf(output_file,"##FILTER=<ID=cov,\"Minimum coverage is %d\">\n",min_coverage);
	    fprintf(output_file,"##FILTER=<ID=p,\"Passes threshold probability of %e\">\n",call_threshold);
	    fprintf(output_file,"##INFO=<ID=NS,\"Number of samples\">\n");	
	    fprintf(output_file,"##FORMAT=<ID=TC,Number=2,Type=Integer,Description=\"Colour read depth +/-\">\n");
	    fprintf(output_file,"##FORMAT=<ID=TL,Number=2,Type=Integer,Description=\"Letter read depth +/-\">\n");
	    fprintf(output_file,"##FORMAT=<ID=GT,Number=1,Type=String,Description=\"Genotype\">\n");
	    fprintf(output_file,"##FORMAT=<ID=DN,Number=1,Type=String,Description=\"Donor name\">\n");
	    fprintf(output_file,"##FORMAT=<ID=P,Number=1,Type=Float,Description=\"Probability Correct\">\n");
	    fprintf(output_file,"##FORMAT=<ID=CS,Number=2,Type=Integer,\"Supporting colour read depth +/-\">\n");
	    fprintf(output_file,"##FORMAT=<ID=LS,Number=2,Type=Integer,\"Supporting letter read depth +/-\">\n");
	    fprintf(output_file,"##FORMAT=<ID=MC,Number=2,Type=Integer,\"Min supporting colour read depth +/-\">\n");
	    fprintf(output_file,"##FORMAT=<ID=ML,Number=2,Type=Integer,\"Min supporting letter read depth +/-\">\n");
	    fprintf(output_file,"#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\t%s\n",donor_name);
    } 

    /* go through each position */
    i = 0;
    int to_travel = (predict_start>0 ? predict_start - hmm_start : 0);
    while(to_travel>0 && i<ref2.size) {
	if (ref2.data[i]!='-') {
		to_travel--;
	}
	i++;
    }
    while(i<ref2.size && ref2.data[i]=='-'){
	i++;
    }
    if (ref2.size<=i) {
	fprintf(stderr,"error: writting VCF output, size of ref2 smaller then index\n");
	exit(1);
    }
    //i=0;
    pos = 1;
    //int end_pos=ref2.size;
    int end_pos=(predict_start > 0 ? predict_end - hmm_start +1 : ref2.size);
    while (i < end_pos) {
	
	int final_pos = startpos + pos - 1 + (predict_start>0 ? predict_start -1 : 0);
        int j;
	for (j=i+1; j<end_pos; j++) {
		//cant be case of -,-,- because cleaned above!
		if (ref2.data[j]==don1.data[j] && ref2.data[j]==don2.data[j]) {
			break;
		}
	}
	if (pos==1 && j<end_pos) {
		j++;
	}
	j--;
	int length = j-i+1;
	char d1[length+1+3]; char d2[length+1+3]; char r[length+1+3];
	int d1_index=0; int d2_index=0; int r_index=0;
	double call_confidence=1.0; double coverage=1.0/0;
	double cs_positive = 0.0; double cs_negative = 0.0;
	double ls_positive = 0.0; double ls_negative = 0.0;
	double min_cs_positive_support=1.0/0.0; double min_cs_negative_support=1.0/0.0;
	double min_ls_positive_support=1.0/0.0; double min_ls_negative_support=1.0/0.0;
	double cs_positive_support=0.0; double cs_negative_support=0.0;
	double ls_positive_support=0.0; double ls_negative_support=0.0;
	for (j=0; j<length; j++) {
		if (don1.data[i+j]=='-' || don2.data[i+j]=='-' || ref2.data[i+j]=='-') {
			break;
		}
	}
	if (j==length) {
		length=1;	
	}
	for (j=0; j<length; j++) {
		if (don1.data[i+j]!='-') {
			d1[d1_index++]=toupper(don1.data[i+j]);
		}
		if (don2.data[i+j]!='-') {
			d2[d2_index++]=toupper(don2.data[i+j]);
		}
		if (ref2.data[i+j]!='-') {
			r[r_index++]=toupper(ref2.data[i+j]);
		}
		call_confidence=MIN(call_confidence,confidence.data[i+j]);
		int z; double current_coverage=0.0;
		for (z=0; z<5; z++) {
			cs_positive+=covmap.colors_positive.data[lookup.data[(i+j)]*5+z];
			cs_negative+=covmap.colors_negative.data[lookup.data[(i+j)]*5+z];
			ls_positive+=covmap.letters_positive.data[lookup.data[(i+j)]*5+z];
			ls_negative+=covmap.letters_negative.data[lookup.data[(i+j)]*5+z];

			current_coverage+=covmap.colors_positive.data[lookup.data[(i+j)]*5+z];
			current_coverage+=covmap.colors_negative.data[lookup.data[(i+j)]*5+z];
			current_coverage+=covmap.letters_positive.data[lookup.data[(i+j)]*5+z];
			current_coverage+=covmap.letters_negative.data[lookup.data[(i+j)]*5+z];
		}
		coverage=MIN(coverage,current_coverage);
	

		double cs_positive_support_current = covmap.colors_positive.data[lookup.data[(i+j)]*5+cmat1.data[i+j]];
		double cs_negative_support_current = covmap.colors_negative.data[lookup.data[(i+j)]*5+cmat1.data[i+j]];	
		if (cmat1.data[i+j]!=cmat2.data[i+j]) {
			cs_positive_support_current+=covmap.colors_positive.data[lookup.data[(i+j)]*5+cmat2.data[i+j]];
			cs_negative_support_current+=covmap.colors_negative.data[lookup.data[(i+j)]*5+cmat2.data[i+j]];
		}
		min_cs_positive_support=MIN(min_cs_positive_support,cs_positive_support_current);
		min_cs_negative_support=MIN(min_cs_negative_support,cs_negative_support_current);
		cs_positive_support+=cs_positive_support_current;
		cs_negative_support+=cs_negative_support_current;
	
		double ls_positive_support_current = covmap.letters_positive.data[lookup.data[(i+j)]*5+lmat1.data[i+j]];
		double ls_negative_support_current = covmap.letters_negative.data[lookup.data[(i+j)]*5+lmat1.data[i+j]];
		if (lmat1.data[i+j]!=lmat2.data[i+j]) {
			ls_positive_support_current+=covmap.letters_positive.data[lookup.data[(i+j)]*5+lmat2.data[i+j]];
			ls_negative_support_current+=covmap.letters_negative.data[lookup.data[(i+j)]*5+lmat2.data[i+j]];
		}
		min_ls_positive_support=MIN(min_ls_positive_support,ls_positive_support_current);
		min_ls_negative_support=MIN(min_ls_negative_support,ls_negative_support_current);
		ls_positive_support+=ls_positive_support_current;
		ls_negative_support+=ls_negative_support_current;
		
	}
	r[r_index]='\0';
	d1[d1_index]='\0';
	if (d1_index==0) {
		sprintf(d1,"D%d",r_index);	
	}
	d2[d2_index]='\0';
	if (d2_index==0) {
		sprintf(d2,"D%d",r_index);	
	}


	char filter_string[256]="PASS";
	int index=0;	
	int failed=0;
	if (min_coverage>coverage) {
		index+=sprintf(filter_string,"cov");
		failed=1;
	}
	if (call_threshold>call_confidence) {
		if (failed==1) {
			index+=sprintf(filter_string+index,";");
		}
		index+=sprintf(filter_string+index,"p");
	}

	char info_string[256];
	//sprintf(info_string,"TC=%d,%d;TL=%d,%d",(int)cs_positive,(int)cs_negative,(int)ls_positive,(int)ls_negative);
	sprintf(info_string,"NS=%d",1);
	double qual_d=-10*(log(1-call_confidence)/log(10));
	qual_d=MIN(4095,qual_d);
	qual_d=MAX(qual_d,0.0);
	int qual = (int)qual_d;

	char format_string[] = "GT:P:TC:TL:MC:ML:CS:LS:DN";
	char donor_format_string[1024];
	sprintf(donor_format_string,"%.2f:%d,%d:%d,%d:%d,%d:%d,%d:%d,%d:%d,%d:%s",call_confidence,
		(int)cs_positive,(int)cs_negative,(int)ls_positive,(int)ls_negative,
		(int)min_cs_positive_support,(int)min_cs_negative_support,(int)min_ls_positive_support,(int)min_ls_negative_support,
		(int)cs_positive_support,(int)cs_negative_support,(int)ls_positive_support,(int)ls_negative_support,donor_name);
	//fprintf(stderr,"%d|%s|%s|%s|%d,%d,%d|%d\n",i,r,d1,d2,ref2.size,don1.size,don2.size,length);
	if (strcmp(d1,r)==0 && strcmp(d2,r)==0) {
		//reference call
		if (report_all == 1) {
			fprintf(output_file,"%s\t%d\t%s\t%s\t%s\t%d\t%s\t%s\t%s",ref_name,final_pos,".",r,".",qual,filter_string,info_string,format_string);
			fprintf(output_file,"\t%s:%s\n","0|0",donor_format_string);
		}
	} else if (strcmp(d1,r)==0) {
		fprintf(output_file,"%s\t%d\t%s\t%s\t%s\t%d\t%s\t%s\t%s",ref_name,final_pos,".",r,d2,qual,filter_string,info_string,format_string);
		fprintf(output_file,"\t%s:%s\n","0|1",donor_format_string);
	} else if (strcmp(d2,r)==0) {
		fprintf(output_file,"%s\t%d\t%s\t%s\t%s\t%d\t%s\t%s\t%s",ref_name,final_pos,".",r,d1,qual,filter_string,info_string,format_string);
		fprintf(output_file,"\t%s:%s\n","1|0",donor_format_string);
        } else if (strcmp(d1,d2)==0) {
		fprintf(output_file,"%s\t%d\t%s\t%s\t%s\t%d\t%s\t%s\t%s",ref_name,final_pos,".",r,d1,qual,filter_string,info_string,format_string);
		fprintf(output_file,"\t%s:%s\n","1|1",donor_format_string);
	} else {
		//triallelic
		fprintf(output_file,"%s\t%d\t%s\t%s\t%s,%s\t%d\t%s\t%s\t%s",ref_name,final_pos,".",r,d1,d2,qual,filter_string,info_string,format_string);
		fprintf(output_file,"\t%s:%s\n","1|2",donor_format_string);
	}

	pos+=r_index;	
	i+=length;

    }    
    
    free(don1.data);
    free(don2.data);
    free(ref2.data);
    free(cmat1.data);
    free(cmat2.data);
    free(lmat1.data);
    free(lmat2.data);
    /*
     * TODO test: there are 8 cases that should be tested in more detail:
     * 4 with no gap in ref - these should be easy
     * 4 more with a gap in ref. (hap1 x hap2 gap on/off). these can be harder
     *   and of multiple lengths...
     */
}
