/*
 * VARiD: A Variation Detection Framework for Colorspace and Letterspace platforms
 * Contact: <varid at cs dot toronto dot edu>
 * Copyright 2010 the authors (see README).
 *
 * This file is part of VARiD.
 *
 * VARiD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * VARiD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VARiD.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "va_log.h"
#include "va_assert.h"
#include "varid.h"
#include "sam2pretty_lib.h"

extern double *cs_error_cnt;
extern double *cs_error_dpts;
extern unsigned int cs_error_size;
extern double *ls_error_cnt;
extern double *ls_error_dpts;
extern unsigned int ls_error_size;

static void addIns(VA_VecI32 ins, VA_VecI32 * insvec );
char *lets = "ACGT";
char *nums = "0123";

long int strchri(char *str, int c)
{
	return (long int)strchr(str, toupper(c));
}

int follows(char prev, char color, char next)
{
	int a, b, c;
	if (!strchri(lets, prev) || !strchri(lets, next)
	    || !strchri(nums, color))
		return 0;
	a = strchri(lets, prev) - (long int)lets;
	b = strchri(lets, next) - (long int)lets;
	c = strchri(nums, color) - (long int)nums;
	if ((a ^ c) == b)
		return 1;
	return 2;
}

void learn_err_mat(char *ls_string, char *cs_string, char *gen_string,
		   int read_start, int read_end, int alignment_length, char* qvals, char* read_name)
{
	//printf("READ:\n%s\n%s\n%s\n",ls_string,cs_string,gen_string);
	int i, x;
	int max_qval=0;
	for (i=0; i<alignment_length; i++) {
		if (qvals[i]>max_qval) {
			max_qval=qvals[i];
		}
	}
	VA_ASSERT(max_qval < MAX_QVAL);
	VA_ASSERT(read_end-read_start<alignment_length);
	//max_qval++; //since we dont use 0, just to keep things sane ? or maybe not :(
	
	if (cs_string) {
		if (max_qval > cs_error_size) {
			VA_ASSERT(1==0);
			/*cs_error_cnt =
			    (double *)realloc(cs_error_cnt,
					      max_qval * sizeof(double));
			cs_error_dpts =
			    (double *)realloc(cs_error_dpts,
					      max_qval * sizeof(double));
			int j;
			for (j = cs_error_size; j < max_qval; j++) {
				cs_error_cnt[j] = cs_error_dpts[j] = 0;
			}
			cs_error_size = max_qval;*/
		}
		/*
		if (read_start == 1) {
			if (x > 1)
				cs_error_cnt[qvals.data[0]]++;
			if (x > 0)
				cs_error_dpts[qvals.data[0]]++;
			if (x>1) {
				fprintf(stderr,"%s\n",read_name);
			}
		}*/
		//fprintf(stderr,"start %d end %d\n",read_start, read_end);
		//fprintf(stderr,"ls %s %d, cs %s %d\n",ls_string,strlen(ls_string), cs_string,strlen(cs_string));
		for (i = 0; i < read_end-read_start+1; i++) {
			if (read_start==1 && i==0) {
				x = follows(cs_string[0], cs_string[0 + 1],
				    ls_string[0]);
			} else {
				x = follows(ls_string[i+read_start-1-1], cs_string[i + read_start -1+1],
				    ls_string[i+read_start-1-1+1]);
			}
			if (x > 1)
				cs_error_cnt[(VA_UINT8)qvals[i]]++;
			if (x > 0)
				cs_error_dpts[(VA_UINT8)qvals[i]]++;
		//	fprintf(stderr,"ls %d, cs %d, ls %d, follows %d, qvals data %d, error cnt / dpts, %f / %f\n",i+read_start-1-1,i+read_start,i+read_start-1-1+1,x,qvals.data[i],cs_error_cnt[qvals.data[i]],cs_error_dpts[qvals.data[i]]);
		}

	} else {
		if (max_qval > ls_error_size) {
			VA_ASSERT(1==0);
			/*ls_error_cnt =
			    (double *)realloc(ls_error_cnt,
					      max_qval * sizeof(double));
			ls_error_dpts =
			    (double *)realloc(ls_error_dpts,
					      max_qval * sizeof(double));
			int j;
			for (j = ls_error_size; j < max_qval; j++) {
				ls_error_cnt[j] = ls_error_dpts[j] = 0;
			}
			ls_error_size = max_qval;*/
		}
		//fprintf(stderr,"%s\n",ls_string);
		for (i =0; i < read_end - read_start + 1; i++) {
			//fprintf(stderr,"%d < %d,%d,%d,re %d\n",qvals.data[i-read_start+1],max_qval,qvals.size,i-read_start+1,read_end);
			if (ls_string[i+read_start-1] != gen_string[i+read_start-1] && 
				gen_string[i+read_start-1]!='-' &&
				ls_string[i+read_start-1] != '-')
				ls_error_cnt[(VA_UINT8)qvals[i]]++;
			ls_error_dpts[(VA_UINT8)qvals[i]]++;
		//	fprintf(stderr,"Comparing position %d, relative %d, letter %c,qval %d, %d\n",i+read_start-1,i,ls_string[i+read_start-1],qvals.data[i],ls_string[i+read_start-1] != gen_string[i+read_start-1] ? 1 : 0);
		}

	}

}

static void getIns(VA_VecU8 chr,
		   VA_UINT32 chrstart, VA_VecI32 * ins, VA_VecI32 * ins_local);

void read_error(char* read_name,int error_number) {
	VA_ASSERT(read_name!=NULL);
	fprintf(stderr,"process_reads.c: %s\n", error_strings[error_number]);
	fprintf(stderr,"offending read is: %s\n",read_name);
	if (drop_bad_reads==0) {
		fprintf(stderr,"please remove read, or use --drop-bad-reads flag\n");
		exit(1);
	} else {
		return;
	}
}

void process_read(char * sam_line,
		  ReadsNode_t * read,
		  VA_VecI32 * insvec, Fasta_t * f, VA_VecU8 * qval_ordering)
{
	//printf("%s\n",sam_line);
	VA_VecU8 read_t;
	char rd_t[BUFSIZ];
	read_t.data = (VA_UINT8 *)rd_t;
	pretty * pa=pretty_from_string(sam_line);
	if (!pa->mapped) {
		fprintf(stderr,"Fatal error! the read %s, is not mapped and passed filtering.\n",pa->read_name);
		exit(1);
	}
	pa->genome_sequence=(char*)f->seq.data;
	pa->genome_length=f->seq.size-1;
	pretty_ls(pa);
	pretty_genome(pa);
	int alignment_length=0;
	if (pa->colour_space) {
		read->isLetterspace=FALSE;
		pretty_cs(pa);
		strcpy((char*)read_t.data,pa->pretty_cs_string);
		read_t.size = strlen((char*)read_t.data);
		alignment_length = read_t.size - 1;
	} else {
		read->isLetterspace=TRUE;
		strcpy((char*)read_t.data,pa->pretty_read_string);
		read_t.size = strlen((char*)read_t.data);
		alignment_length = read_t.size;
	}
	pretty_qualities(pa);
	int read_start=pa->pretty_clipped_read_start;
	int read_end=pa->pretty_clipped_read_end;
	int alignment_chrstart=MIN(pa->genome_start_unpadded,pa->genome_end_unpadded);
	int alignment_chrend=MAX(pa->genome_start_unpadded,pa->genome_end_unpadded);
	//fprintf(stderr,"\n%s\naln starts %d, aln ends %d\n",pa->read_name,alignment_chrstart,alignment_chrend);
	//fprintf(stderr,"align len %d\n",alignment_length);
	//fprintf(stderr,"re: %d , rs: %d\n",read_end,read_start);
	if (alignment_chrend>pa->genome_length) {
		read_error(pa->read_name,0);
		return;
	}
	read->strand=pa->reverse ? '-' : '+';
	read->name=pa->read_name;
	pa->read_name=NULL; //tell it not to be freed at the end
	char readstring[BUFSIZ];
	strcpy(readstring,(char*)read_t.data);
	char genome_string[BUFSIZ];
	strcpy(genome_string,pa->pretty_genome_string);


	int i;
	char qvals[BUFSIZ];
	strcpy(qvals,pa->pretty_read_qualities);
	//fprintf(stderr,"%s\n %s\n %s\n",readstring,genome_string,qvals);
	if (qvals[0]=='*' && qvals[1]=='\0') {
		for (i=0; i<alignment_length; i++) {
			if (read_t.data[i]=='-') {
				qvals[i]=GAP_QVAL;
			} else {
				qvals[i]= (i+1) >= (MAX_QVAL-1) ? MAX_QVAL-2 : (i+1);
			}
			qval_ordering->data[(VA_UINT8)(qvals[i])]=1;
		}
		qvals[i]='\0';
	} else {
		if (strlen(qvals)!=alignment_length) {
			fprintf(stderr,"%s, %d %d\n",sam_line,strlen(qvals),alignment_length);
			read_error(pa->read_name,1);
			return;	
		}
		for (i=0; i<alignment_length; i++) {
			if (qvals[i]==' ') {
				qvals[i]=GAP_QVAL;
			}
			qval_ordering->data[(VA_UINT8)(qvals[i])]=1;
		}
	}
	
	
	if (strlen(qvals)!=alignment_length) {
		read_error(pa->read_name,1);
		return;	
	}

	VA_VecU8 ls_chr = {0 , 0};
	VA_VecU8 cs_chr = {0 , 0};
		
	//Messy but using existing func, get the complement if need be
	if (read->strand=='-') {
		VA_VecU8 chr_vec; chr_vec.size=strlen(genome_string); chr_vec.data=(VA_UINT8*)genome_string;
		VA_VecU8 temp = get_complement(chr_vec);
		memcpy(genome_string,temp.data,strlen(genome_string));
		free(temp.data);
	}
	
	/* deal with the read */
	VA_VecU8 temp;
	if (!pa->colour_space) {
		//put the read into number space
		for (i=0; i<read_t.size; i++) {
			read_t.data[i]=tolower(read_t.data[i]);
		}
		read_t = nuc2nuc(read_t);

		//extract only the read base pairs and gaps inbetween them
		//that we care about
		int read_index=0;
		int bp_to_use=read_end-read_start+1;
		VA_ASSERT(bp_to_use>0);
		for (i=read_start-1; i<read_t.size && bp_to_use>0; i++) {
			read_t.data[read_index++]=read_t.data[i];
			bp_to_use--;
		}
		read_t.size=read_index;
		if (bp_to_use!=0) {
			read_error(read->name,2);
			free(read_t.data);
			return;	
		}

		//set the ls_chr to correct value on the STACK!
		ls_chr.data = (VA_UINT8*)genome_string + read_start - 1; ls_chr.size = read_t.size;
		read->ls_qvals = va_allocVecU8(read_t.size);
		memcpy(read->ls_qvals.data,qvals+read_start-1,read->ls_qvals.size);
		//If it's on the negative strand, deal with it
		if (read->strand == '-') {
			//flip the qvals
			fliplrU8(read->ls_qvals);
			//flip the read, and get complement
			fliplrU8(read_t);
			VA_VecU8 complement=get_complement(read_t);
			free(read_t.data);
			read_t = complement;
			//flip the genome string
			//it is already complemented - see above
			fliplrU8(ls_chr);
		}
		read->ls_seq = read_t;
		read->ls_chrstart=alignment_chrstart;
		read->ls_chrend=alignment_chrend;
	} else {
		/* color-space is more complicated */
		VA_VecU8 letter_read;
		VA_VecU8 first_letter;

		//translate the first letter to numbers
		read_t.data[0]=tolower(read_t.data[0]);
		temp.data = read_t.data;
		temp.size = 1;
		first_letter = nuc2nuc(temp);

		/* translate to */
		for (i = 1; i < read_t.size; i++) {
			if (read_t.data[i] == '-')
				read_t.data[i] = '4';
			if (read_t.data[i] == '.')
				read_t.data[i] = '5';
			read_t.data[i-1] = read_t.data[i] - '0';
		}
		read_t.size--;

		//for (i=0; i<read_t.size; i++) {
		//	printf("%d",read_t.data[i]);
		//}
		//printf("\n");

		//push the letter 1 in
		temp.data = read_t.data;
		temp.size = read_start;
		/* translate through the beginning of the read */
		letter_read = col2nuc(first_letter.data[0], temp);

		//col2nuc includes the first letter in the colspace read in return
		first_letter.data[0] = letter_read.data[read_start];
		free(letter_read.data);
	
		//already translated through first number/letter so start
		//at read_start+1 'th element, which is index read_start
		int read_index=0;
		int bp_to_use=read_end-read_start+1-1; //one bp already in letter
		//printf("going to use %d bp\n",bp_to_use);
		if (bp_to_use==0) {
			read_error(read->name,3);
			return;	
		}
		for (i=read_start-1+1; i<read_t.size && bp_to_use>0; i++) {
			read_t.data[read_index++]=read_t.data[i];
			bp_to_use--;
		}
		read_t.size=read_index;
		if (bp_to_use!=0) {
			read_error(read->name,4);
			return;	
		}
		//for (i=0; i<read_t.size; i++) {
		//	printf("%d",read_t.data[i]);
		//}
		//printf("\n");

		//get the first letter
		//printf("read size %d\n",read_t.size);
		if (read->strand=='+') {
			read->ls_seq = first_letter;
			read->ls_chrstart = read->ls_chrend = alignment_chrstart; 
		} else {
			//get the complement of the read
			read->ls_seq = get_complement(first_letter);
			free(first_letter.data);
			//get the first letter
			//ls_chr.data = (VA_UINT8*)genome_string + read_end-1; ls_chr.size=1;
			read->ls_chrstart = read->ls_chrend = alignment_chrend;
		}
		//printf("start %d, end %d\n",alignment_chrstart,alignment_chrend);
		ls_chr.data = (VA_UINT8*)genome_string + read_start-1; ls_chr.size=1;
		read->ls_qvals = va_allocVecU8(1);
		read->ls_qvals.data[0] = qvals[read_start-1];

		//check if the read has a insert at first position
		/*if (ls_chr.data[0]=='-') {
			read->ls_chrend--;
			read->cs_chrstart = alignment_chrstart;
		} else {
			read->cs_chrstart = alignment_chrstart+1;
		}
		read->cs_chrend = alignment_chrend;*/
		//allocate and set up qvals
		read->cs_qvals = va_allocVecU8(read_t.size);
		memcpy(read->cs_qvals.data,qvals+read_start-1+1,read->cs_qvals.size);


		int begins_with_insert=( ls_chr.data[0]=='-' ? 1 : 0);
		if (begins_with_insert) {
			read->ls_chrend--;
		}
		int ends_with_insert=( genome_string[read_end-1]=='-' ? 1 : 0);
		//printf("%s\n",genome_string+read_end-1);
		//printf("%s\n",genome_string);
		//flip the cs stuff on negative, push numbers one right
		//printf("read s %d, read e %d\n",read_start,read_end);
		//printf("%s\n",ls_chr.data[0]=='-' ? "ls is gap" : "ls is not gap");
		//printf("%s\n",ends_with_insert==1 ? "ends with insert" : "no t end with ins");
		if (read->strand == '+') {
			//printf("positive strand\n");
			read->cs_chrstart=alignment_chrstart+1 - begins_with_insert;
			read->cs_chrend=alignment_chrend;
			cs_chr.data = (VA_UINT8*)genome_string+(read_start-1)+1;
			cs_chr.size=read_end-read_start+1-1;
		} else {
			//printf("neg. strand\n");
			read->cs_chrstart=alignment_chrstart+1 -ends_with_insert;
			read->cs_chrend=alignment_chrend-1 + 1;
			//have the genome string already shifted by one, since its transistions
			//going in wrong direction
			//R:  A A C G       T T G C
			//  T 3 0 1 3       0 1 3 3 G
			//    A0 1 3         0 1 3C
			//    A 0 1 3         0 1 3	
			//                        C
			cs_chr.data = (VA_UINT8*)genome_string+(read_start-1)+1-1;
			cs_chr.size=read_end-read_start+1-1;
			//printf("%d\n%s\n",cs_chr.size, cs_chr.data);
			//flip the read
			fliplrU8(read_t);	
			//flip the quality values for colour space
			fliplrU8(read->cs_qvals);
			/* if there are any gaps, they have to be slightly shifted */
			int i;
			for (i = 1; i < read_t.size; i++) {
				if ((read_t.data[i] == 4) && (i > 0)) {
					VA_UINT8 t = read_t.data[i - 1];
					read_t.data[i - 1] = read_t.data[i];
					read_t.data[i] = t;
				}
			}
			//ls_chr and cs_chr is already complemented
			fliplrU8(cs_chr);
			ls_chr.data+=cs_chr.size-1;
			//printf("%d\n%s\n",cs_chr.size, cs_chr.data);
		}
		//for (i=0; i<read_t.size; i++) {
		//	printf("%d",read_t.data[i]);
		//}
		//printf("\n%s\n",cs_chr.data);
		//TODO, loss of letter space signal!
		if (read->ls_seq.data[0]==4) {
			read->ls_seq.size=0;
			free(read->ls_seq.data);
			read->ls_seq.data=NULL;
			read->ls_chrstart=0; read->ls_chrend=0;
			read->ls_qvals.size=0;	
			free(read->ls_qvals.data);
			read->ls_qvals.data=NULL;
		}
		read->cs_seq = va_allocVecU8(read_t.size);
		int x; 
		for (x=0; x<read_t.size; x++) {
			VA_ASSERT(read_t.data[x]<6);
		}
		memcpy(read->cs_seq.data, read_t.data,
		       read_t.size);
		//VA_ASSERT(read->ls_seq.data[0]<5);	
	}
	if (read->ls_seq.size != read->ls_qvals.size) {
		read_error(read->name,5);
		return;
	}
	if (read->cs_seq.size != read->cs_qvals.size) {
		read_error(read->name,6);
		return;
	}
	
	
	/*printf("Genome string:\n%s\n",pa->pretty_genome_string);
	printf("Read string:\n%s\n",readstring);
	printf("First letter %d\n",read->ls_seq.data[0]);*/
	
	VA_VecI32 ins;

	ins.size=0; ins.data=NULL;
	getIns(ls_chr, read->ls_chrstart, &ins, &read->ls_ins);
	addIns(ins,insvec);

	ins.size=0; ins.data=NULL;
	getIns(cs_chr, read->cs_chrstart, &ins, &read->cs_ins);
	addIns(ins,insvec);

	/* another field for old times' sake */
	read->extras.haplome = 1;

	if (load_learned_file == NULL) {
		if (!pa->colour_space) {
			learn_err_mat(readstring, NULL, pa->pretty_genome_string, read_start,
				      read_end, alignment_length,qvals,read->name);
		} else {
			char tline[BUFSIZ];
			strcpy(tline,pa->pretty_read_string);
			learn_err_mat(tline, readstring, pa->pretty_genome_string, read_start,
				      read_end, alignment_length,qvals,read->name);
		}
	}

	//print read stats
	if (read_alignments_file!=NULL) {
		fprintf(read_alignments_file,"Read:\n%s\n",readstring);
		fprintf(read_alignments_file,"Genome:\n%s\n",pa->pretty_genome_string);
		fprintf(read_alignments_file,"Read start: %d\tRead end: %d\n",read_start,read_end);
		for (i=0; i<read->ls_seq.size; i++) {
			fprintf(read_alignments_file,"%d",read->ls_seq.data[i]);
		}
		fprintf(read_alignments_file,"\n");
		for (i=0; i<read->cs_seq.size; i++) {
			fprintf(read_alignments_file,"%d",read->cs_seq.data[i]);
		}
		fprintf(read_alignments_file,"\n");
		fprintf(read_alignments_file,"LS : Chr start: %d\tChr end: %d\n",read->ls_chrstart, read->ls_chrend);
		fprintf(read_alignments_file,"CS : Chr start: %d\tChr end: %d\n",read->cs_chrstart, read->cs_chrend);
		fprintf(read_alignments_file,"Inserts (1based):\n");
		for (i=0; i<insvec->size; i++) {
			fprintf(read_alignments_file,"%d,",insvec->data[i]);
		}
		fprintf(read_alignments_file,"\nLS Read Inserts (1based):\n");
		for (i=0; i<read->ls_ins.size; i++) {
			fprintf(read_alignments_file,"%d,",read->ls_ins.data[i]);
		}
		fprintf(read_alignments_file,"\nCS Read Inserts (1based):\n");
		for (i=0; i<read->cs_ins.size; i++) {
			fprintf(read_alignments_file,"%d,",read->cs_ins.data[i]);
		}
		fprintf(read_alignments_file,"\nLS Read Qvals:\n");
		for (i=0; i<read->ls_qvals.size; i++) {
			fprintf(read_alignments_file,"%d,",read->ls_qvals.data[i]);
		}
		fprintf(read_alignments_file,"\nCS Read Qvals:\n");
		for (i=0; i<read->cs_qvals.size; i++) {
			fprintf(read_alignments_file,"%d,",read->cs_qvals.data[i]);
		}
		fprintf(read_alignments_file,"\n\n-------------------------\n");
	}
	pretty_free(pa);
}


static void addIns(VA_VecI32 ins, VA_VecI32 * insvec ) {

	if (ins.size > 0) {
		VA_VecI32 u = va_unique(ins);
		int j;
		int i;
		for (i = 0; i < u.size; i++) {
			int counti = 0;
			int s = 0;

			for (j = 0; j < ins.size; j++)
				if (u.data[i] == ins.data[j])
					counti++;

			for (j = 0; j < insvec->size; j++)
				if (u.data[i] == insvec->data[j])
					s++;

			if (s < counti) {
				size_t oldsz = insvec->size;

				if (oldsz > 0)
					*insvec = va_reallocVecI32(*insvec,
							     oldsz + counti -
							     s);
				else
					*insvec = va_allocVecI32(counti - s);

				for (j = 0; j < counti - s; j++)
					insvec->data[oldsz + j] = u.data[i];
			}
		}
		free(ins.data);
	}
}
static void getIns(VA_VecU8 chr,
		   VA_UINT32 chrstart, VA_VecI32 * ins, VA_VecI32 * ins_local)
{

	VA_ASSERT(ins != NULL);
	VA_ASSERT(ins_local != NULL);

	memset(ins, 0, sizeof(VA_VecI32));
	memset(ins_local, 0, sizeof(VA_VecI32));

	int n = 0; int i;
	for (i = 0; i < chr.size; i++)
		if (chr.data[i] == '-')
			n++;

	if (n > 0) {
		*ins = va_allocVecI32(n);
		*ins_local = va_allocVecI32(n);

		n = 0;
		for (i = 0; i < chr.size; i++) {
			if (chr.data[i] == '-') {
				ins_local->data[n] = i-n;
				ins->data[n] =
				    chrstart + ins_local->data[n] - 1;
				n++;
			}
		}
	}
}
